import os
import xml.etree.ElementTree as ET
from typing import TypeVar, Set, Tuple, IO

from graphviz import Digraph

T = TypeVar('T')
class NBA():

    def __init__(self,
                 states: Set[int],
                 alphabet: Set[T],
                 tr: Set[Tuple[int, T, int]],
                 init: Set[int],
                 final: Set[int]) -> None:
        self.states = states
        self.alphabet = alphabet
        self.tr = tr
        self.init = init
        self.final = final

    def print_nba_graph(self) -> Digraph:
        graph = Digraph()

        # Add states
        for state in self.states:
            graph.node(str(state), shape='circle')
            if state in self.init:
                graph.node(str(state), style='bold')
            if state in self.final:
                graph.node(str(state), peripheries='2')

        # Add transitions
        for transition in self.tr:
            from_state, read, to_state = transition
            graph.edge(str(from_state), str(to_state), label=str(read))

        return graph

def nba_parse_from_goal(stream: IO) -> NBA:
    """
    Parse a buchi automaton from the xml format of GOAL.
    """
    # Parse the XML file
    tree = ET.parse(stream)
    root = tree.getroot()

    # Extract necessary information
    states = set()
    alphabet = set()
    tr = set()
    init = set()
    final = set()

    for state in root.findall('./stateset/state'):
        states.add(int(state.attrib['sid']))

    for symbol in root.findall('./alphabet/prop'):
        alphabet.add(symbol.text)
    for symbol in root.findall('./alphabet/proposition'):
        alphabet.add(symbol.text)


    for transition in root.findall('./transitionset/transition'):
        from_state = int(transition.find('from').text)
        try:
            read = transition.find('read').text
        except:
            read = transition.find('label').text
        to_state = int(transition.find('to').text)
        tr.add((from_state, read, to_state))

    for state_id in root.findall('./initialstateset/stateid'):
        init.add(int(state_id.text))

    for state_id in root.findall('./acc/stateid'):
        final.add(int(state_id.text))

    # Create NBA instance
    nba = NBA(states, alphabet, tr, init, final)
    return nba

directory = os.getcwd()
for filename in os.listdir(directory):
    if filename.endswith("gff.xml"):
        filepath = os.path.join(directory, filename)
        with open(filepath, 'r') as f:

            nba = parse_xml(f)

            # Print NBA instance for verification
            print(filename)
            # print("NBA states:", nba.states)
            # print("NBA alphabet:", nba.alphabet)
            # print("NBA transitions:", nba.tr)
            # print("NBA initial states:", nba.init)
            # print("NBA final states:", nba.final)
            print(nba.print_nba_graph())
            print()

