""" Non-Deterministic Buchi Automaton (NBA)

A NBA is the tuple: <states, alphabet, tr, init, final>

A state is an integer number of a tuple of states (inductive definition)
The alphabet is a finite set of symbols
A transition relation tr is a tuple in State x Action x State
init is a subset of states
final is a subset of states

Example of use:

# Creates the states
s1 = (1,)
s2 = (2,)

# Creates the NBA that accept the language "a (b)^omega"
t = NBA(set([s1, s2]),
        set(["a","b"]),
        set([(s1,"a",s2),(s2,"b",s2)]),
        set([s1]),
        set([s2]))
"""

from __future__ import annotations
from collections.abc import Mapping, Iterator, Iterable

from typing import (
    TypeAlias, Type, Tuple, Callable, Optional, NewType,
    List, Set,
    Union, TypedDict, IO, TypeVar
)

from mc308.data_structures.ts import States, Prop, Trans

from functools import reduce, partial
import itertools

import xml.etree.ElementTree as ET


T = TypeVar('T')
class NBA():

    def __init__(self : NBA,
                 states : frozenset[States],
                 alphabet : frozenset[T],
                 tr : frozenset[Trans],
                 init : frozenset[States],
                 final : frozenset[States]) -> None:
        self.states = set([a for a in states])
        self.alphabet = set([a for a in alphabet])
        self.tr = set([t for t in tr])
        self.init = set([i for i in init])
        self.final = set([i for i in final])

        self.nextmap = {s : set() for s in self.states}
        for (src, a, dst) in self.tr:
            self.nextmap[src].add((a,dst))




    def is_empty(self : NBA) -> Tuple[True,Option[List[State]]]:
        """
        Check if the language accepted by the NBA is empty and, if not,
        returns an accepting run.

        The function returns:
        - (True, None)    if the language(self) is empty
        - (False, run)    otherwise

        run must represents an INFINITE RUN of the NBA.

        How can we represent that?

        Every run of the NBA can be represented implicitly with a
        finite sequence of states:

        s0 s1 ... s-(i-1) s-i ... s-(i+k) s-

        where s-i is an accepting state.
        The above sequence represents the INFINITE run of the NBA:

        s0 s1 ... s-(i-1) (s-i s-(i+1) ... s-(i+k))^(\omega)

        Where:
          - s0 s1 ... s-(i-1) is a finite prefix (it may also be of zero length)
          - (s-i s-(i+1) ... s-(i+k))^(\omega) is the part of the run repeats infinitely often.


        HINT: implement the double DFS algorithm (or compute the reachable SCCs, ...)
        """

        raise NotImplementedError


    def is_run(self : NBA, run : List[States], accepting : bool) -> Tuple[bool,Option[List[T]]]:
        print("is_run is still not implemented!")

        return (False, None)


def nba_parse_from_goal(stream: IO) -> NBA:
    """
    Parse a buchi automaton from the xml format of GOAL.
    """
    # Parse the XML file
    tree = ET.parse(stream)
    root = tree.getroot()

    # Extract necessary information
    states = set()
    alphabet = set()
    tr = set()
    init = set()
    final = set()

    for state in root.findall('./stateset/state'):
        states.add(int(state.attrib['sid']))

    for symbol in root.findall('./alphabet/prop'):
        alphabet.add(symbol.text)
    for symbol in root.findall('./alphabet/proposition'):
        alphabet.add(symbol.text)


    for transition in root.findall('./transitionset/transition'):
        from_state = int(transition.find('from').text)
        try:
            read = transition.find('read').text
        except:
            read = transition.find('label').text
        to_state = int(transition.find('to').text)
        tr.add((from_state, read, to_state))

    for state_id in root.findall('./initialstateset/stateid'):
        init.add(int(state_id.text))

    for state_id in root.findall('./acc/stateid'):
        final.add(int(state_id.text))

    # Create NBA instance
    nba = NBA(states, alphabet, tr, init, final)
    return nba
